//====================================================================
// SDL_bmf.h
//
// Copyright (c) 2000 Joseph A. Knapka.
//
// License: GPL/LGPL, as appropriate.
//
// This file defines the SDL_BmapFont class, which loads fonts
// described by XML font-descriptions (possibly loading auxiliary
// bitmap files), and renders scaled, colored text using the font,
// returning transparent, display-formatted SDL_Surfaces that can be
// blitted wherever the user desires. It can also render text directly
// to an existing SDL_Surface, and can convert between BMP and pure
// XML representations, allowing fonts to be edited with any text
// editor.
//
// Example:
//
//    SDL_BmapFont f1("font1.xml",32,0xFF,0xFF,0x00));
//    SDL_Surface* rendered_text = f1.DrawText("Hello, world!");
//
// This code loads the font described in the file "font1.xml" (format
// of which is described below), scales it if necessary to obtain
// 32-pixel-tall characters, then constructs an SDL_Surface with
// transparent background containing the text "Hello, world!" in the
// selected color (red=FF,green=FF,blue=00). (If you use multiple
// colors and/or sizes of the same font, multiple SDL_BmapFont objects
// can share the underlying master font bitmap.)
//
// The format of the XML font description can be any one of the
// following:
//
// <FONT bmapfile="font.bmp" w="8" h="16" columns="32" rows="8"/>
//
// This tag describes a bitmapped font stored in a separate file.  In
// this case, the character bitmaps will be found in the BMP file
// "font.bmp", in a regular array of 8x16-pixel cells; the character
// bitmap is 32 characters wide and 8 characters tall. The characters
// map onto their ASCII values in row order; thus character code 0
// will be rendered using the bitmap in cell (0,0), character code 16
// will be rendered using the bitmap in cell (16,0), and character
// code 51 will be rendered using the bitmap in cell (19,1). When the
// font is loaded, an SDL_Surface describing the entire font is
// created from "font.bmp". When a particular character is used, an
// SDL_Surface for that character and color is created and cached in
// scaled form, so that it can be immediately blitted thenceforth.
//
// <FONT bmapfile="font.bmp">
// <CH c="112" x="100" y="0" w="8" h="16" d="3"/>
// ...additional <CH(aracter)/> tags...
// </FONT>
//
// This tag also describes a bitmapped font stored in a separate file.
// In this case the XML description denotes the character locations
// within "font.bmp" on a character-by-character basis using <CH>
// tags.  This allows characters of varying widths and heights to be
// represented.  The "d" (descender) attribute indicates that the
// bitmap for the character in question should be displaced down by
// the indicated number of pixels when displayed; this can help ensure
// that character baselines line up properly when variable-height
// characters are used. The caching process is as described above.
//
// <FONT_DESCRIPTION>
// <CH c="65" w="5" h="7" d="0">
// -----
// --**-
// -*--*
// -****
// -*--*
// -*--*
// -*--*
// </CH>
// ...additional <CH> tags...
// </FONT>
//
// In this case, there is no external bitmap file; the character
// bitmaps are encoded as text within the <CH> tags. When this font is
// loaded, an SDL_Surface containing the entire font is created from
// the <CH> tags, after which caching proceeds as described above.
//
// The <FONT_DESCRIPTION> tag need not be stored in an external file;
// it can be passed directly to the SDL_BmapFont contructor.
//
// External bitmap files are expected to be white-on-black BMP-format
// files, though any BMP with a solid black backround will work.
//
// SDL_BmapFont objects can also save themselves as either pure XML
// text containing text character bitmaps, or as BMP files with
// separate XML descriptions suitable for inclusion in source code or
// in a resource file.
//
//====================================================================
#ifndef SDL_BMAPFONT__
#define SDL_BMAPFONT__

#include <SDL.h>
#include <expat++.h>

#define XML_FONT_TAG (&("FONT")[0])
#define XML_CHARACTER_TAG (&("CH")[0])
#define XML_FONT_TYPE_ARG (&("type")[0])
#define XML_FONT_CONSTANT (&("constant")[0])
#define XML_FONT_DESCRIBED (&("description")[0])
#define XML_FONT_ENCODED (&("encoded")[0])
#define XML_CHARCODE_ARG (&("c")[0])
#define XML_WIDTH_ARG (&("w")[0])
#define XML_HEIGHT_ARG (&("h")[0])
#define XML_ROWS_ARG (&("rows")[0])
#define XML_COLUMNS_ARG (&("columns")[0])
#define XML_XCOORD_ARG (&("x")[0])
#define XML_YCOORD_ARG (&("y")[0])
#define XML_DESCENDER_ARG (&("d")[0])
#define XML_FILE_ARG (&("bmapfile")[0])
#define XML_BMAP_SET '*'
#define XML_BMAP_CLR '-'

class SDL_BmapFont {
public:

  // Construct a font object as in the example above.
  SDL_BmapFont(const char* font_desc_or_filename,
	       Uint16 req_height,
	       Uint8 red,Uint8 blue,Uint8 grn);

  // Construct a font object using an existing surface as
  // the master bitmap. This can be used to allow multiple SDL_BmapFont
  // objects with different colors and/or scaling factors to
  // share a master bitmap.
  SDL_BmapFont(SDL_Surface* master_bmap,
	       const char* fonst_desc_or_filename,
	       Uint16 req_height,
	       Uint8 red,Uint8 blue,Uint8 grn);

  // Construct a font object using the given SDL_BmapFont's master
  // bitmap as the bitmap.
  SDL_BmapFont(const SDL_BmapFont& orig,
	       Uint16 req_height,
	       Uint8 red,Uint8 blue,Uint8 grn);

  // Construct a font object using a backing bitmap instead of
  // a single solid color. If bbmap == 0, then the master bitmap
  // will be used *as is*, with no colorization done at all.
  // In that case, any solid-black pixel will still be rendered
  // transparent, however.
  SDL_BmapFont(const char* font_desc_or_filename,
	       Uint16 req_height,
	       SDL_Surface* bbmap);

  // Construct a font object using an existing surface as
  // the master bitmap. This can be used to allow multiple SDL_BmapFont
  // objects with different colors and/or scaling factors to
  // share a master bitmap.
  SDL_BmapFont(SDL_Surface* master_bmap,
	       const char* fonst_desc_or_filename,
	       Uint16 req_height,
	       SDL_Surface* bbmap);

  // Construct a font object using the given SDL_BmapFont's master
  // bitmap as the bitmap.
  SDL_BmapFont(const SDL_BmapFont& orig,
	       Uint16 req_height,
	       SDL_Surface* bbmap);

  ~SDL_BmapFont();

  // Call this after constructing the object. It returns true iff
  // the font can be used for rendering.
  bool Usable() { return usable_; }

  // Call this to get a descriptive error string if Usable() fails.
  char* GetErrorDesc() { return (char*)(errstr_.c_str()); }

  // D.Finkler:
  // Render font and cache all chars in font
  void CacheFont();

  // D.Finkler:
  // Render the given text and cache all chars in text
  void CacheText(const char* text);

  // Render the given text and return a transparent SDL_Surface
  // containing the resulting bitmap. hdiv is the horizontal pixel
  // gap between characters. The resulting surface will be display-
  // formatted.
  SDL_Surface* DrawText(const char* text,int hdiv=0);

  // Render the given text directly onto the specified surface
  // at the given coordinates.
  void DrawText(SDL_Surface* dest,int x,int y,const char* text,int hdiv=0);

  // Prepare the characters in the given string for rendering. This
  // is not necessary, but can speed up later DrawText() calls by
  // avoiding the character-rendering overhead. Returns the number
  // of characters successfully rendered and cached.
  int PrepareCharacters(const char* str);

  // Save the font as XML text. Return true on success, false
  // on failure.
  bool SaveAsXML(const char* xmlfile);

  // Save as a BMP file with auxiliary XML description.
  bool SaveAsBMPAndXML(const char* bmapfile,const char* xmlfile);

  // Get the maximum height of any character in the font.
  int GetMaxHeight();

  // D.Finkler:
  int GetCharWidth(const unsigned char ch);

  // Find the dimensions of the region necessary to hold the
  // rendered string.
  void GetTextDimension(const char* txt,Uint16* ht,Uint16* wd,int hgap=0);

private:

  std::string fdesc_;
  SDL_Surface* font_bmap_; // Bitmap containing entire font.
  bool rel_font_bmap_; // True if we must release the font_bmap_.
  Uint16 req_ht_; // Requested height.
  float scale_; // The chosen scaling factor.
  Uint16 max_ht_; // Maximum scaled height.
  Uint8 r_,g_,b_;
  char colorkey_b_; // Blue to use in colorkey -- (b_==0)?0f:00
  SDL_Surface* bbmap_; // Backing bitmap.
  bool no_color_; // If true, just scale, don't colorize.

  bool usable_;
  LXMLString errstr_;

  // Struct used to cache the colored, scaled bitmap of a single
  // character:
  struct CharStruct {
    // The following specify the location in the master bitmap.
    Uint16 x;
    Uint16 y;
    Uint16 wd;
    Uint16 ht;
    // The descender is the scaled vertical distance to displace the character
    // when rendering text.
    Uint16 desc;
    // Scaled character representation.
    SDL_Surface* bmap;
    CharStruct(Uint16 x_,Uint16 y_,Uint16 wd_,Uint16 ht_,Uint16 desc_=0,SDL_Surface* bmap_=0) :
      x(x_),y(y_),wd(wd_),ht(ht_),desc(desc_),bmap(bmap_) {
    }
  };

  // Map of char codes to CharStructs.
  std::map<int,CharStruct*> char_bmaps_;

  void load_bmap_font();

  // Scale and render a particular character given its cache description.
  SDL_Surface* scale_and_render_char(CharStruct& cs);

  // Get/set pixels in an SDL_Surface. get_pixel() returns true iff the
  // selected pixel is not black.
  void set_pixel(SDL_Surface* image,int x,int y,Uint8 r,Uint8 g,Uint8 b);
  bool get_pixel(SDL_Surface* image,int x,int y,Uint8* r,Uint8* g,Uint8* b);

  // Load an XML font description. This produces the master font
  // bmap for font_bmap_, and also extracts the cache descriptors into char_bmaps_.
  // It either loads the master bitmap from a file, or synthesizes it from
  // the per-character bitmap data contained in the XML font description.
  bool loadXmlFont(const XMLTag* font);

  // Three distictly different code paths, depending on the font descriptor.
  bool loadConstantSzFont(const XMLTag* font);
  bool loadXMLDescribedFont(const XMLTag* font);
  bool loadXMLEncodedFont(const XMLTag* font);

  // "Expanded bitmap" - character bitmap encoded as a text string.)
  class XBmap {
    XMLString bits;
    long wd;
    long ht;
    long sz;
  public:
    bool getPixel(int x,int y) {
      if (y*wd+x>sz) return false;
      return bits[y*wd+x] == XML_BMAP_SET;
    }
    int getWd() { return wd; }
    int getHt() { return ht; }
    XBmap(const XMLTag* ch);
  };

  // Utilities to operate on the XML font description.
  int fontTotalWidth(const XMLTag* font);
  int fontMaxHeight(const XMLTag* font);
  XBmap charTagToExBmap(const XMLTag* ch);
  void exBmapToSurface(XBmap& xbmap,SDL_Surface* img,int x,int y);
  bool tryToLoadBmapFile(const XMLTag* font);
  SDL_Surface* makeBmapFromXMLChars(const XMLTag* font);
  void choose_scale_factor(const XMLTag* font);
  bool extract_char_descs(const XMLTag* font,bool make_bmap);
  void gen_const_descs(int wd,int ht,int rows,int cols);
  bool extract_one_desc(const XMLTag* ch,int& atx,bool make_bmap);

  // Utilities to save XML fonts.
  std::ostream& saveCharToStream(std::ostream& os,int code,CharStruct* cs,bool desc_only=false);
};

#endif // SDL_BMAPFONT__
