#include "expat++.h"

#include <algorithm>
#include <memory>
#include <string.h>
#include <strstream>

using namespace std;

///////////////////// Methods for XMLText /////////////////

std::basic_string<XML_Char> XMLText::GetText() const { return text_; }

///////////////////// Methods for XMLParser /////////////////

// Construct the parser, register callbacks.
XMLParser::XMLParser() : cur_tag_(0)
{
  parser_ = XML_ParserCreate(0);
  XML_SetElementHandler(parser_,
			&XMLParser::startElementHandler,
			&XMLParser::endElementHandler);
  XML_SetCharacterDataHandler(parser_,
			      charDataHandler);
  XML_SetUserData(parser_,(void*)this);
}

void XMLParser::startElementHandler(void* parser,
				    const XML_Char* name,
				    const XML_Char** attrs)
{
  ((XMLParser*)parser)->startHandler(name,attrs);
}

void XMLParser::endElementHandler(void* parser,
				  const XML_Char* name)
{
  ((XMLParser*)parser)->endHandler(name);
}

void XMLParser::charDataHandler(void* parser,
				const XML_Char* text,
				int len)
{
  ((XMLParser*)parser)->charHandler(text,len);
}

void XMLParser::startHandler(const XML_Char* name,
			     const XML_Char** attrs)
{
  // If we have a cur_tag_, put it on the stack and
  // create a new one.
  if (cur_tag_) tag_stack_.push_back(cur_tag_);
  cur_tag_ = new XMLTag(name);

  // Handle the attributes.
  int at = 0;
  while (attrs[at]) {
    cur_tag_->add_attr(attrs[at],attrs[at+1]);
    at += 2;
  }
}

void XMLParser::endHandler(const XML_Char* name)
{
  // If there is a tag on the stack, add the current tag
  // as a child, then pop the stack and install the top
  // as cur_tag_.
  if (!tag_stack_.empty()) {
    XMLTag* t = tag_stack_.back();
    tag_stack_.pop_back();
    t->add_element(cur_tag_);
    cur_tag_ = t;
  }
}

void XMLParser::charHandler(const XML_Char* text,
			    int len)
{
  // If no cur_tag_, this sucks. Just throw it away.
  // Otherwise, make a new text element add add it to
  // cur_tag_.
  if (cur_tag_) {
    std::basic_string<XML_Char> st(text,len);
    bool added = false;
    if (cur_tag_->NElements()>0) {
      const XMLElement* el = cur_tag_->NthElement(cur_tag_->NElements()-1);
      if (el->GetType() == XMLElement::TEXT) {
	XMLText* t = (XMLText*)el;
	t->add_text(st);
	added = true;
      }
    }
    if (!added) {
      XMLText* t = new XMLText(st);
      cur_tag_->add_element(t);
    }
  }
}

// Parse an XML string out of a stream. The first character
// in the stream must be a tag-start character "<".
XMLTag* XMLParser::Parse(istream& in,
			 XML_Error& err,
			 long& errloc,
			 std::basic_string<XML_LChar>& errstr)
{
  errstr = "";
  XMLTag* result = Parse(in);
  if (!result) {
    // We aborted before reading all input!
    cleanup_tag_stack();
    err = XML_GetErrorCode(parser_);
    errstr = XML_ErrorString(err);
    errloc = XML_GetCurrentByteIndex(parser_);
  }
  return result;
}

XMLTag* XMLParser::Parse(istream& in,
			 XML_Error& err,
			 long& errline,
			 long& errcol,
			 std::basic_string<XML_LChar>& errstr)
{
  errstr = "";
  XMLTag* result = Parse(in);
  if (!result) {
    // We aborted before reading all input!
    cleanup_tag_stack();
    err = XML_GetErrorCode(parser_);
    errstr = XML_ErrorString(err);
    errline = XML_GetCurrentLineNumber(parser_);
    errcol = XML_GetCurrentColumnNumber(parser_);
  }
  return result;
}

XMLTag* XMLParser::Parse(istream& in)
{
  cur_tag_ = 0;

  // All we really need to do here is read blocks from <in>
  // and feed them to this->parser_. The callback handlers
  // take care of the rest.

  char buf[1024];
  while (!in.eof()) {
    std::fill(buf,buf+1024,0x00);
    in.read(buf,1023);
    int final = 0;
    int len = strlen(buf);
    if (len < 1023) final = 1;
    if (!XML_Parse(parser_,buf,len,final)) {
      break;
    }

    if (cur_tag_ == (void*)0xffffffff) cout << "Something really bad happened..." << endl;

  }

  if (!in.eof()) {
    return 0;
  }

  if (cur_tag_ == (void*)0xffffffff) cout << "Something really bad 2 happened..." << endl;
  return cur_tag_;
}


void XMLParser::cleanup_tag_stack()
{
  if (tag_stack_.size() > 0) {
    XMLTag* top = tag_stack_.front();
    delete top;
    while (tag_stack_.size() > 0) tag_stack_.pop_back();
  }
}

///////////////// Methods for XMLTag ////////////////////////

XMLTag::~XMLTag()
{
  // We must delete all our children.
  int ne = NElements();
  for (int el=0; el<ne; ++el) delete elements_[ne];
}

std::vector< std::basic_string<XML_Char> > XMLTag::AttrNames() const
{
  std::map< std::basic_string<XML_Char>,std::basic_string<XML_Char> >::
    const_iterator it = attributes_.begin();

  std::vector< std::basic_string<XML_Char> > result;
  while (it != attributes_.end()) {
    result.push_back((*it).first);
    ++it;
  }
  return result;
}

// Retrieve the nth contained element. Return 0 if no such.
const XMLElement* XMLTag::NthElement(unsigned int n) const
{
  if (n<elements_.size()) return elements_[n];
  else return 0;
}

// Find a subtag with name and attributes matching tag_template,
// which is just an empty tag with the given characteristics
// (for exmaple, <MY_TAG my_attr="excellent!"/>). Attributes
// not specified in the template are ignored in the search.
const XMLTag* XMLTag::FindMatchingSubtag(XML_Char* tag_template) const
{
  XMLParser par;
  XML_Error err;
  long errloc;
  basic_string<XML_LChar> errstr;
  istrstream in(tag_template);
  auto_ptr<XMLTag> temp_tag(par.Parse(in,err,errloc,errstr));
  if (!temp_tag.get()) {
    return 0;
  }

  basic_string<XML_Char> temp_name = temp_tag->GetName();
  vector<basic_string<XML_Char> > temp_attrs = temp_tag->AttrNames();

  unsigned int n = NElements();
  for (unsigned int el=0; el<n; ++el) {
    const XMLElement* check_el = NthElement(el);
    if (check_el->GetType() != TAG) continue;
    const XMLTag* check_tag = (XMLTag*)check_el;
    if (check_tag->GetName() == temp_name) {
      bool match = true;
      for (int aa=0; aa<temp_attrs.size(); ++aa) {
	basic_string<XML_Char> attr = temp_attrs[aa];
	basic_string<XML_Char> temp_val;
	temp_tag->AttrValue(attr,&temp_val);
	basic_string<XML_Char> val;
	if (check_tag->AttrValue(attr,&val) != XMLTag::ATTR_SUCCESS) {
	  match = false;
	  break;
	}
	if (val != temp_val) {
	  match = false;
	  break;
	}
      }
      if (match) {
	return check_tag;
      }
    }
  }
  return 0;
}

// Retrieve attribute values in various ways. Each method
// attempts to convert the named attribute's value into
// the appropriate type and assign the value to the argument.
// If this succeeds, the second argument receives the value
// and ATTR_SUCCESS is returned. If the conversion fails, then
// ATTR_TYPE_ERROR is returned. If no attribute with the given
// name is found, ATTR_NONE is returned.

XMLTag::ATTR_RC XMLTag::AttrValue(const std::basic_string<XML_Char>& attr,
				  std::basic_string<XML_Char>* result) const
{
  *result = "";
  std::map<std::basic_string<XML_Char>,std::basic_string<XML_Char> >::
    const_iterator it = attributes_.find(attr);
  if (it == attributes_.end()) return ATTR_NONE;
  *result = (*it).second;
  return ATTR_SUCCESS;
}

XMLTag::ATTR_RC XMLTag::AttrValue(const std::basic_string<XML_Char>& attr,
				  long* result) const
{
  std::map<std::basic_string<XML_Char>,std::basic_string<XML_Char> >::
    const_iterator it = attributes_.find(attr);
  if (it == attributes_.end()) return ATTR_NONE;
  istrstream in((*it).second.c_str());
  in >> *result;
  if (in.fail()) return ATTR_TYPE_ERROR;
  return ATTR_SUCCESS;
}

XMLTag::ATTR_RC XMLTag::AttrValue(const std::basic_string<XML_Char>& attr,
				  double* result) const
{
  std::map<std::basic_string<XML_Char>,std::basic_string<XML_Char> >::
    const_iterator it = attributes_.find(attr);
  if (it == attributes_.end()) return ATTR_NONE;
  istrstream in((*it).second.c_str());
  in >> *result;
  if (in.fail()) return ATTR_TYPE_ERROR;
  return ATTR_SUCCESS;
}

XMLText::~XMLText() {}

XMLElement::~XMLElement() {}
