//=========================================================
// expat++.h
//
// This is a C++ wrapper around James Clark's "expat"
// XML parsing library. It builds an in-memory parse tree
// that can be used to extract tag, text, and attribute
// data. It is intended mainly for small, frequently-
// accessed XML documents, such as ones that describe
// application configuration data, for example.
//
// The parse tree is represented by a pair of concrete
// classes, XMLTag and XMLText. XMLText represents
// an unmarked text string, while XMLTag represents a
// tag, along with its attributes and contents. The
// contents of an XMLTag is just a list of the XMLText
// and XMLTag objects within the tag, in left-to-right
// order. You can iterate over the contents by using the
// XMLTag::NElements() and XMLTag::NthElement()
// methods. You can search a tag for a subtag matching
// a template tag using the XMLTag::FindMatchingSubtag()
// method. You can find all a tag's attribute names
// using the XMLTag::AttrNames() method. You can find the
// value of a particular attribute by using the
// XMLTag::AttrValue() methods.
//
// To parse a document, do the following:
//
//     ifstream infile("filename.xml");
//     std::string parse_error;
//     XMLParser parser;
//     XMLTag* top_tag = parser.Parse(infile,&parse_error);
//     // Do stuff with top_tag...
//     delete top_tag;
//
// If non-0, top_tag is now a pointer to a tag representing
// the parsed document; otherwise, parse_error contains
// the error string returned by expat.
//
// LICENSE: FIX ME!!!
//
// Copyright (c) 2000 Joseph A. Knapka
//=========================================================
#ifndef EXPAT_PLUSPLUS_H__
#define EXPAT_PLUSPLUS_H__

#include <xmlparse.h>
#include <string>
#include <iostream>
#include <map>
#include <vector>

typedef std::basic_string<XML_Char> XMLString;
typedef std::basic_string<XML_LChar> LXMLString;

// The XMLElement class is the common base for the XMLText
// and XMLTag classes. It merely provides a way to discover
// whether an element is a tag or unmarked text.
class XMLElement {
public:
  enum TYPE { TEXT, TAG };

  XMLElement(TYPE t) : type_(t) {}
  virtual ~XMLElement() = 0;

  TYPE GetType() const { return type_; }

private:

  const TYPE type_;

};

// The XMLText class describes an unmarked text string
// with a document.
class XMLText : public XMLElement {
public:

  std::basic_string<XML_Char> GetText() const;

  ~XMLText();

private:

  friend class XMLParser;

  std::basic_string<XML_Char> text_;

  XMLText(const std::basic_string<XML_Char>& text) :
    XMLElement(XMLElement::TEXT), text_(text) {} 

  // Used to merge adjacent text elements.
  void add_text(const std::basic_string<XML_Char>& text) { text_ += text; }

};

// The XMLTag class describes a tag and its contents,
// including subtags and text.
class XMLTag : public XMLElement {
public:

  // Get the tag name.
  std::basic_string<XML_Char> GetName() const { return name_; }

  // Find the number of elsments immediately contained within
  // this tag.
  unsigned int NElements() const { return (unsigned int)elements_.size(); }

  // Retrieve the nth contained element. Return 0 if no such.
  const XMLElement* NthElement(unsigned int n) const;

  // Find a subtag with name and attributes matching tag_template,
  // which is just an empty tag with the given characteristics
  // (for exmaple, <MY_TAG my_attr="excellent!"/>). Attributes
  // not specified in the template are ignored in the search.
  const XMLTag* FindMatchingSubtag(XML_Char* tag_template) const;

  // Retrieve the list of attribute names.
  std::vector< std::basic_string<XML_Char> > AttrNames() const;

  // Retrieve attribute values in various ways. Each method
  // attempts to convert the named attribute's value into
  // the appropriate type and assign the value to the argument.
  // If this succeeds, the second argument receives the value
  // and ATTR_SUCCESS is returned. If the conversion fails, then
  // ATTR_TYPE_ERROR is returned. If no attribute with the given
  // name is found, ATTR_NONE is returned.

  enum ATTR_RC { ATTR_SUCCESS=0, ATTR_TYPE_ERROR=-1, ATTR_NONE=-2 };

  ATTR_RC AttrValue(const std::basic_string<XML_Char>& attr,
		    std::basic_string<XML_Char>* result) const;
  ATTR_RC AttrValue(const std::basic_string<XML_Char>& attr,
		    long* result) const;
  ATTR_RC AttrValue(const std::basic_string<XML_Char>& attr,
		    double* result) const;

  ~XMLTag();

private:

  friend class XMLParser;

  std::basic_string<XML_Char> name_;

  XMLTag(const XML_Char* name) :
    XMLElement(XMLElement::TAG), name_(name) {}

  // Add a sub-element.
  void add_element(XMLElement* el) { elements_.push_back(el); }

  // Add an attribute.
  void add_attr(const XML_Char* name,const XML_Char* value) {
    attributes_[name] = value;
  }

  // Contents.
  std::vector<XMLElement*> elements_;

  // Attributes.
  std::map<std::basic_string<XML_Char>,std::basic_string<XML_Char> >
  attributes_;

};

class XMLParser {
public:

  // Construct the parser, register callbacks.
  XMLParser();

  // Parse an XML string out of a stream. The first character
  // in the stream must be a tag-start character "<".
  XMLTag* Parse(std::istream& in,
		XML_Error& err,
		long& errloc,
		std::basic_string<XML_LChar>& errstr);
  XMLTag* Parse(std::istream& in,
		XML_Error& err,
		long& errline,
		long& errcol,
		std::basic_string<XML_LChar>& errstr);
  XMLTag* Parse(std::istream& in);

private:

  // Handlers we give to expat.
  static void startElementHandler(void* parser,
				  const XML_Char* name,
				  const XML_Char** attrs);
  static void endElementHandler(void* parser,
				const XML_Char* name);
  static void charDataHandler(void* parser,
			      const XML_Char* text,
			      int len);

  // Member versions of above.
  void startHandler(const XML_Char* name,
		    const XML_Char** attrs);
  void endHandler(const XML_Char* name);
  void charHandler(const XML_Char* text,
		   int len);

  XML_Parser parser_;
  XMLTag* cur_tag_;
  std::vector<XMLTag*> tag_stack_;

  void cleanup_tag_stack();
  
};


#endif // EXPAT_PLUSPLUS_H__
